#include <stdio.h>
#include <mpi.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <memory.h>

#define TAG 1

int main(int argc, char ** argv) {

//Two arguments here, the first, N, is the size of the array.
//N should be a multiple of world_size for this first attempt....
//The second is the range, R, of the random numbers.

	int my_rank, world_size;
	int my_sum = 0;
	float total_sum = 0;
	float my_avg;
	float total_avg;

        char message[256], name[80];
        MPI_Status status;

	int j, k;
        int N = atoi(argv[1]);
	int R = atoi(argv[2]);

	int seed;
	double rand1 = 0.0;
	double temp = 0.0;
	seed = time(NULL);
	srand(seed);

	MPI_Init(&argc, &argv); /* note that argc and argv are passed by address */

        MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
        MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	int SubN = N/world_size;

	int *RandomArray = NULL; // this will be the array for the random numbers, created on node 0
	int *SubArray = NULL; // where the smaller pieces of the array go to each node
	float *SubAvgs = NULL; // the array that node0 will store the individual averages in
	int *SendCounts = NULL; // for computing the number of elements sent to each node
	int localSendCount = 0; // how many elements each array will get
	int *displacer = NULL; // offset for scatterv for array sending

	if (my_rank == 0)
	{
		SendCounts = (int *)malloc(world_size * sizeof(int));
		displacer = (int *)malloc(world_size * sizeof(int));
		SubAvgs = (float *)malloc(world_size * sizeof(float));
		RandomArray = (int *)malloc(N * sizeof(int));

		printf("Node %d: N is %d, SubN is %d. Now populating the random array. \n", my_rank, N, SubN);

		for(k=0; k<world_size-1; k++)
		{
			displacer[k] = k*SubN;
			SendCounts[k] = SubN;
			printf("Node %d: Send count for node %d is %d, starting at entry %d. \n", my_rank, k, SendCounts[k], displacer[k]);
		}

		displacer[world_size-1] = (world_size-1)*SubN;
		SendCounts[world_size-1] = N - (world_size - 1)*SubN;
		printf("Node %d: Send count for node %d is %d, starting at entry %d. \n", my_rank, world_size-1, SendCounts[world_size-1], displacer[world_size-1]);

		for(k=0; k<N; k++)
		{
			RandomArray[k] = k;
//			rand1 = rand();
//			temp = rand1 / (double) RAND_MAX;
//			RandomArray[k] = R*temp;
//			RandomArray[k] = -R + 2*R*temp;
		}

		printf("Node %d: First entry is %d, last entry is %d. \n", my_rank, RandomArray[0], RandomArray[N-1]);

	}

	// we send the number of elements to average to each node now....

	MPI_Scatter(SendCounts, 1, MPI_INT, &localSendCount, 1, MPI_INT, 0, MPI_COMM_WORLD);

	printf("Node %d: My local send count is %d. \n", my_rank, localSendCount);

	SubArray = (int *)malloc(localSendCount * sizeof(int));


	// initializing the SubArray for each node, putting known nontrivial values at each ends
	SubArray[0] = -2;
	SubArray[localSendCount-1] = -1;
	for(k=1; k < localSendCount-1; k++)
	{
		SubArray[k] = 0;
	}

	printf("Node %d: localSendCount is %d, entry 0 is %d, entry %d is %d. \n", my_rank, localSendCount, SubArray[0], localSendCount-1, SubArray[localSendCount-1]);

	MPI_Scatterv(RandomArray, SendCounts, displacer, MPI_INT, SubArray, localSendCount, MPI_INT, 0, MPI_COMM_WORLD);
//	MPI_Scatter(RandomArray, SubN, MPI_INT, SubArray, SubN, MPI_INT, 0, MPI_COMM_WORLD);

	printf("Node %d: Received data. Array size  is %d, entry 0 is %d, entry %d is %d. \n", my_rank, localSendCount, SubArray[0], localSendCount-1, SubArray[localSendCount-1]);

	for(k=0; k<localSendCount; k++)
	{
		my_sum = my_sum + SubArray[k];
	}

	my_avg = my_sum/(float)localSendCount;
	printf("Node %d: My SubArray sum is %d, and the average is %4.4f \n", my_rank, my_sum, my_avg);

	MPI_Gather(&my_avg, 1, MPI_FLOAT, SubAvgs, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);

	if (my_rank == 0)
	{

		for(k=0; k<world_size; k++)
		{
			printf("Node %d: Average from node %d is %4.4f \n", my_rank, k, SubAvgs[k]);
			total_sum = total_sum + SubAvgs[k];
		}

		total_avg = total_sum/(float)world_size;
		printf("Node %d: Total average is %4.4f \n", my_rank, total_avg);
	}

    MPI_Finalize();

}
